"""Houses Option"""

from typing import Dict, Any, Iterable
from dataclasses import dataclass

@dataclass
class Option:
    """Object Type"""
    
    Name: str
    Value: Any
    
    def as_dict(self) -> Dict[str, Any]:
        """Gets the object as it's dict representation"""

        return {self.Name: self.Value}

    #def __iter__(self) -> Iterable:
    #    return iter(self.as_dict().items())



"""Houses Supports Put Method"""

from sys import version_info
from typing import TypeVar

if version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.request.supports_types._supports_methods import SupportsMethods
from pyrestsdk.type.model import BaseEntity

S = TypeVar("S", bound=BaseEntity)


class SupportsPutMethod(SupportsMethods):
    """Supports Put Method Type"""

    def Put(self, input_object: S) -> Self:
        """Sets request to put request"""

        self._update_request_type(HttpsMethod.PUT, input_object)

        return self



"""Houses Supports Get Method"""

from sys import version_info

if version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.request.supports_types._supports_methods import SupportsMethods


class SupportsGetMethod(SupportsMethods):
    """Supports Get Method Type"""

    @property
    def Get(self) -> Self:
        """Sets request to get request"""

        self._update_request_type(HttpsMethod.GET, None)

        return self



"""House Abstract Basic Credential"""

from abc import abstractmethod
from pyrestsdk.credential._abstract_credential import AbstractCredential


class AbstractBasicCredential(AbstractCredential):
    """Abstract Basic Credential Type"""

    @abstractmethod
    def get_basic(self, /) -> str:
        """Gets the basic credential encoded string"""



from __future__ import annotations
from typing import Protocol, Optional, Union, List, TypeVar, Type
from abc import abstractmethod
from pyrestsdk.type.model import BaseEntity

T = TypeVar("T", bound=BaseEntity)


class SupportsBaseInvokeRequest(Protocol):
    """Supports Base Invoke Request Type"""

    @property
    @abstractmethod
    def generic_type(self) -> Type[T]:
        ...

    @property
    @abstractmethod
    def input_object(self) -> T:
        ...

    @abstractmethod
    def Send(self) -> Optional[Union[List[T], T]]:
        ...

    @abstractmethod
    def invoke_request(self) -> Optional[Union[List[T], T]]:
        """Invokes the prepared request"""



"""Houses Base Request Builder"""

from __future__ import annotations
from typing import TypeVar
from pyrestsdk import AbstractServiceClient
from pyrestsdk.requestbuilder._abstract_request_builder import AbstractRequestBuilder


S = TypeVar("S", bound=AbstractServiceClient)
B = TypeVar("B", bound="BaseRequestBuilder")


class BaseRequestBuilder(AbstractRequestBuilder):
    """Base Request Builder Type"""



"""Houses Abstract Request Builder"""

from __future__ import annotations
from typing import (
    TypeVar,
    Generic,
    Type,
    final,
    get_args,
)
from pyrestsdk import AbstractServiceClient
from pyrestsdk.type.model import (
    QueryOption,
    HeaderOption,
)

B = TypeVar("B", bound="AbstractRequestBuilder")
S = TypeVar("S", bound=AbstractServiceClient)
O = TypeVar("O", QueryOption, HeaderOption)
T = TypeVar("T")


class AbstractRequestBuilder(Generic[T]):
    """Abstract Request Type"""

    _client: S
    _request_url: str

    def __init__(self: B, request_url: str, client: S) -> None:

        super().__init__()

        self._request_url: str = request_url
        self._client = client

    @property
    def Client(self: B) -> S:
        """Gets the Client"""

        return self._client

    @property
    @final
    def generic_type(self: B) -> Type[T]:
        """Gets the the type argument provided"""

        # used if type arg is provided in constructor
        orig_value = getattr(self, "__orig_class__", None)

        if orig_value is None:
            # used if typ arg is provided when subclassing
            orig_bases = getattr(self, "__orig_bases__")
            orig_value = orig_bases[0]

        _type: Type[T] = get_args(orig_value)[0]

        return _type

    @property
    def request_url(self: B) -> str:
        """Gets/Sets the request URL

        Returns:
            str: The request URL
        """

        return self._request_url

    @request_url.setter
    def request_url(self: B, value: str) -> None:
        self._request_url = value

    def append_segment_to_request_url(self, url_segment: str) -> str:
        """Gets a URL that is the request builder's request URL with the segment appended.

        Args:
            url_segment (str): The segment to append to the request URL.

        Returns:
            str: A URL that is the request builder's request URL with the segment appended.
        """

        if not url_segment.startswith("/"):
            url_segment = f"/{url_segment}"

        return f"{self.request_url}{url_segment}"



"""Houses access token"""

from pyrestsdk.type.model.token._base_token import BaseToken


class AccessToken(BaseToken):
    """Access Token Type"""

    expires_on: int


# AccessToken.expires_on.__doc__ = """The token's expiration time in Unix time."""



"""Supports Types"""

from pyrestsdk.request.supports_types._supports_base_invoke_request import (
    SupportsBaseInvokeRequest,
)
from pyrestsdk.request.supports_types._supports_invoke_collection_request import (
    SupportsInvokeCollectionRequest,
)
from pyrestsdk.request.supports_types._supports_invoke_request import (
    SupportsInvokeRequest,
)

from pyrestsdk.request.supports_types._supports_delete_method import (
    SupportsDeleteMethod,
)
from pyrestsdk.request.supports_types._supports_get_method import SupportsGetMethod
from pyrestsdk.request.supports_types._supports_post_method import SupportsPostMethod
from pyrestsdk.request.supports_types._supports_put_method import SupportsPutMethod

__all__ = [
    "SupportsInvokeRequest",
    "SupportsBaseInvokeRequest",
    "SupportsInvokeCollectionRequest",
    "SupportsDeleteMethod",
    "SupportsGetMethod",
    "SupportsPostMethod",
    "SupportsPutMethod",
]



from typing import TypeVar
from typing import (
    final,
    TypeVar,
    List,
    Union,
    Type,
    Optional,
    Iterable,
    get_args,
)
from abc import abstractmethod
import logging
from urllib.parse import urlparse
from requests import Response
from pyrestsdk import AbstractServiceClient
from pyrestsdk.request._abstract_request import AbstractRequest
from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.type.model import (
    BaseEntity,
    QueryOption,
    HeaderOption,
    HeaderOptionCollection,
    QueryOptionCollection,
)
from pyrestsdk.request._abstract_request import AbstractRequest

T = TypeVar("T")
B = TypeVar("B", bound="Request")
O = TypeVar("O", QueryOption, HeaderOption)
S = TypeVar("S", bound=AbstractServiceClient)

Logger = logging.getLogger(__name__)


class Request(AbstractRequest[T]):

    _method: HttpsMethod
    _request_url: str
    _query_options: QueryOptionCollection
    _header_options: HeaderOptionCollection

    def __init__(
        self: B, request_url: str, client: S, options: Optional[Iterable[O]]
    ) -> None:

        super().__init__(request_url, client)

        self._method: HttpsMethod = HttpsMethod.GET
        self._request_url: str = self._initialize_url(request_url)
        self._query_options: QueryOptionCollection = QueryOptionCollection()
        self._header_options: HeaderOptionCollection = HeaderOptionCollection()
        self._parse_options(options)
    
    @property
    def header_options(self) -> HeaderOptionCollection:
        """Gets the headers"""
        return self._header_options

    @property
    def request_method(self) -> HttpsMethod:
        """Gets/Sets the https method"""
        return self._method

    @request_method.setter
    def request_method(self, value: HttpsMethod) -> None:
        self._method = value
        Logger.info("%s.Method: _method set to %s", type(self).__name__, value)

    @property
    def request_url(self: B) -> str:
        """Gets/Sets the request URL

        Returns:
            str: The request URL
        """

        return self._request_url

    @request_url.setter
    def request_url(self: B, value: str) -> None:
        self._request_url = value

    @property
    @final
    def generic_type(self: B) -> Type[T]:

        # used if type arg is provided in constructor
        orig_value = getattr(self, "__orig_class__", None)

        if orig_value is None:
            # used if typ arg is provided when subclassing
            orig_bases = getattr(self, "__orig_bases__")
            orig_value = orig_bases[0]

        _type: Type[T] = get_args(orig_value)[0]

        return _type

    @property
    def Client(self: B) -> S:
        """Gets the Client"""

        return self._client

    def _initialize_url(self, request_url: str) -> str:
        """Parses the query parameters from URL"""

        Logger.info(f"%s._initialize_url: function called", type(self).__name__)

        if not request_url:
            pass

        url = urlparse(request_url)

        if url.query:
            query_string = url.query
            query_options = query_string.split("&")
            for option in query_options:
                query_parameter, value = option.split("=")
                _query_parameter = QueryOption(query_parameter, value)
                self.query_options.append(_query_parameter)

        return url._replace(query="").geturl()

    def Send(self, __object: Optional[T]) -> Optional[Union[List[T], T]]:
        """Submits the request and returns the expected return"""

        Logger.info("%s.Send: method called", type(self).__name__)

        return self.send_request(__object)

    def send_request(self, value: Optional[T]) -> Optional[Union[List[T], T]]:
        """Makes the desired request and returns the desired return type"""

        Logger.info("%s.SendRequest: method called", type(self).__name__)

        _response = self._send_request(value)

        if _response is None:
            return None

        return self.parse_response(_response)

    @abstractmethod
    def parse_response(
        self, _response: Optional[Response]
    ) -> Optional[Union[List[T], T]]:
        """Parses the response into the expected return"""

    def append_segment_to_request_url(self, url_segment: str) -> str:
        """Gets a URL that is the request builder's request URL with the segment appended.

        Args:
            url_segment (str): The segment to append to the request URL.

        Returns:
            str: A URL that is the request builder's request URL with the segment appended.
        """

        if not url_segment.startswith("/"):
            url_segment = f"/{url_segment}"

        return f"{self.request_url}{url_segment}"



"""Houses Entity Request Builder"""

from __future__ import annotations
from typing import TypeVar, Iterable, Optional
from abc import abstractmethod
from pyrestsdk import AbstractServiceClient
from pyrestsdk.requestbuilder._base_request_builder import BaseRequestBuilder
from pyrestsdk.request import BaseRequest
from pyrestsdk.type.model import Option

S = TypeVar("S", bound=AbstractServiceClient)
B = TypeVar("B", bound=BaseRequestBuilder)
R = TypeVar("R", bound=BaseRequest)
O = TypeVar("O", bound=Option)


class EntityRequestBuilder(BaseRequestBuilder):
    """Entity Request Builder Type"""

    @property
    @abstractmethod
    def request(self: B) -> R:
        """Gets request with no options"""

    @abstractmethod
    def Request(self: B, options: Optional[Iterable[O]]) -> R:
        """Gets request"""
